#include	<windows.h>
#include	"common.h"
#include	"pc_core.h"
#include	"i286.h"
#include	"iobridge.h"
#include	"memory.h"
#include	"bios.h"
#include	"dosio.h"
#include	"fdc.h"
#include	"fddfile.h"
#include	"fdd_mtr.h"
#include	"milstr.h"
#include	"trace.h"

	char	hd_name[2][MAX_PATH];
	char	fdmode = 0;

enum {
	HDDTYPE_SASI,
	HDDTYPE_SCSI,
	HDDTYPE_NONE,
};

typedef struct {
	char	vhd[3];
	char	ver[4];
	char	delimita;
	char	comment[128];
	DWORD	padding;
	WORD	mbsize;
	WORD	sectorsize;
	BYTE	sectors;
	BYTE	surfaces;
	WORD	tracks;
	long	totals;
} V98SCSI_TABLE;

typedef struct {
	long	totals;
	WORD	tracks;
	WORD	size;
	BYTE	sectors;
	BYTE	surfaces;
	WORD	type;
	int		headersize;
	char	fname[MAX_PATH];
} SXSI_TABLE;

typedef struct {
	DWORD x00;
	DWORD x04;
	DWORD header_size;
	DWORD image_size;
	DWORD sector_size;
	DWORD sector;
	DWORD head;
	DWORD track;
} HDI_HEADER;

static SXSI_TABLE defsasi = {615*33*8, 615, 256, 33, 8,
											HDDTYPE_SASI, 256, {0x00}};
static SXSI_TABLE defscsi = {40*16*32*8, 40*16, 256, 32, 8,
											HDDTYPE_SCSI, 220, {0x00}};

SXSI_TABLE	sxsi_hd[4];


BYTE work[65536];

static BYTE nosyscode[] = {
			0xfc,0x8c,0xc8,0x8e,0xd8,0xb8,0x00,0xa0,0x8e,0xc0,0xb9,0x00,
			0x10,0xb8,0x20,0x00,0x33,0xff,0xf3,0xab,0xb0,0xe1,0xb9,0xe0,
			0x1f,0xaa,0x47,0xe2,0xfc,0xbe,0x2d,0x00,0x33,0xff,0xac,0x3c,
			0x00,0x74,0x04,0xaa,0x47,0xeb,0xf7,0xeb,0xfe,0x4e,0x6f,0x20,
			0x53,0x79,0x73,0x74,0x65,0x6d,0x20,0x46,0x69,0x6c,0x65,0x73,
			0x2e,0x00};


static	BYTE	mtr_c = 0;
static	BYTE	mtr_r = 0;

void bios0x1b_init(void) { }


// -------------------------------------------------------------------- FDD

static void init_fdd_equip(void) {

	MEM_DISK_EQUIP &= 0x0f00;
	MEM_DISK_EQUIP |= (WORD)(~fdmode) & 3;
	MEM_DISK_EQUIP |= (WORD)fdmode << 12;
}

static void biosfd_setchrn(void) {

	fdc.C = I286_CL;
	fdc.H = I286_DH;
	fdc.R = I286_DL;
	fdc.N = I286_CH;
}

static void biosfd_resultout(void) {

	BYTE	*p = &mem[0x00564 + fdc.us*8];
	*p++ = (fdc.hd << 2) | fdc.us;
	*p++ = 0;
	*p++ = 0;
	*p++ = fdc.C;
	*p++ = fdc.H;
	*p++ = fdc.R;
	*p++ = fdc.N;
	*p++ = fdc.ncn;
}

static BOOL biosfd_seek(BYTE track) {

	fdc.ncn = track;
	mtr_c = track;
	if (fdd_seek()) {
		return(FAILURE);
	}
	return(SUCCESS);
}

static BYTE fdd_operate(BYTE type) {

	BYTE	ret_ah = 0x60;
	WORD	size;
	WORD	pos;
	WORD	accesssize;
	WORD	secsize;

	mtr_c = 0xff;
	mtr_r = 0;

	if ((I286_AH & 0x0f) != 0x0a) {
		fdc.crcn = 0;
	}
	if (I286_AL & 0x0c) {
		return(0x60);
	}
	if ((I286_AH & 0x0f) != 0x03) {
		CTRL_FDMEDIA = type;
		if (fdc.us != (I286_AL & 0x03)) {
			fdc.us = I286_AL & 0x03;
			fdc.crcn = 0;
		}
		if (fdc.hd != (I286_DH & 0x01)) {
			fdc.hd = I286_DH & 1;
			fdc.crcn = 0;
		}
		if (!fdd_diskready(fdc.us)) {
			return(0x60);
		}
	}

	switch(I286_AH & 0x0f) {
		case 0x00:								// V[N
			if ((I286_AH & 0x10) && (biosfd_seek(I286_CL))) {
				ret_ah = 0xe0;
				break;
			}
			ret_ah = 0x00;
			break;

		case 0x01:								// xt@C
			ret_ah = 0x60;
			break;

		case 0x03:								// 
			init_fdd_equip();
			ret_ah = 0x00;
			break;

		case 0x04:								// ZX
			if (fdd_diskaccess()) {
				ret_ah = 0xe0;
			}
			else {
				ret_ah = 0x01;
				if (fdd_diskprotect(fdc.us)) {
					ret_ah |= 0x10;
				}
				if (type == DISKTYPE_2DD) {
					ret_ah |= 4;
				}
			}
			break;

		case 0x05:								// f[^̏
#if 1
			if ((I286_AH & 0x10) && (biosfd_seek(I286_CL))) {
				ret_ah = 0xe0;
				break;
			}
			biosfd_setchrn();
			if (fdd_diskprotect(fdc.us)) {
				ret_ah = 0x70;
				break;
			}
			if (fdc.N < 8) {
				secsize = 128 << fdc.N;
			}
			else {
				secsize = 128 << 8;
			}
			size = I286_BX;
			pos = I286_BP;
			while(size) {
				if (size > secsize) {
					accesssize = secsize;
				}
				else {
					accesssize = size;
				}
				i286_memstr_read(I286_ES, pos, fdc.buf, accesssize);
				if (fdd_write()) {
					if ((I286_AH & 0x80) && (!fdc.hd)) {
						fdc.hd = 1;
						fdc.H = 1;
						fdc.R = 1;
						if (!biosfd_seek(fdc.treg[fdc.us])) {
							continue;
						}
					}
					break;
				}
				pos += accesssize;
				size -= accesssize;
				fdc.R++;
				mtr_r++;
			}
			if (!size) {
				ret_ah = 0x00;
			}
			else {
				ret_ah = 0xc0;
			}
#else
			ret_ah = 0x70;
#endif
			break;

		case 0x02:								// ffׂ̈̓ǂݍ
		case 0x06:								// f[^̓ǂݍ
			if ((I286_AH & 0x10) && (biosfd_seek(I286_CL))) {
				ret_ah = 0xe0;
				break;
			}
			biosfd_setchrn();
			if (fdc.N < 8) {
				secsize = 128 << fdc.N;
			}
			else {
				secsize = 128 << 8;
			}
			size = I286_BX;
			pos = I286_BP;
			while(size) {
				if (size > secsize) {
					accesssize = secsize;
				}
				else {
					accesssize = size;
				}
				if (fdd_read()) {
					if ((I286_AH & 0x80) && (!fdc.hd)) {
						fdc.hd = 1;
						fdc.H = 1;
						fdc.R = 1;
						if (!biosfd_seek(fdc.treg[fdc.us])) {
							continue;
						}
					}
					break;
				}
				i286_memstr_write(I286_ES, pos, fdc.buf, accesssize);
				pos += accesssize;
				size -= accesssize;
				fdc.R++;
				mtr_r++;
			}
			if (!size) {
				ret_ah = 0x00;
			}
			else {
				ret_ah = 0xc0;
			}
			break;

		case 0x07:						// V_OփV[N
			if (biosfd_seek(0)) {
				ret_ah = 0xe0;
				break;
			}
			ret_ah = 0x00;
			break;

		case 0x0a:						// READ ID
			fdc.hd = I286_DH & 1;
			if (I286_AH & 0x10) {
				if (biosfd_seek(I286_CL)) {
					ret_ah = 0xe0;
					break;
				}
			}
			if (fdd_readid()) {
				ret_ah = 0xa0;
				break;
			}
			mtr_r++;
			ret_ah = 0x00;
			I286_CL = fdc.C;
			I286_DH = fdc.H;
			I286_DL = fdc.R;
			I286_CH = fdc.N;
			break;

		case 0x0d:						// tH[}bg
			fdc.hd = I286_DH & 1;
			if (I286_AH & 0x10) {
				if (biosfd_seek(I286_CL)) {
					ret_ah = 0xe0;
					break;
				}
			}
			if (fdd_diskprotect(fdc.us)) {
				ret_ah = 0x70;
				break;
			}
			fdc.d = I286_DL;

			fdc.N = 3;
			fdc.sc = 8;
//			ret_ah = 0x70;
			ret_ah = 0x00;
			break;
	}
	biosfd_resultout();
#if 1
	fddmtr_seek(fdc.us, mtr_c, mtr_r);
	I286_IP = 0x40;
#endif
	return(ret_ah);
}


// -------------------------------------------------------------------- SASI

static void init_sasi_equip(void) {

	BYTE	i;
	WORD	bit;

	MEM_DISK_EQUIP &= 0xf0ff;
	for (i=0, bit=0x0100; i<2; i++, bit<<=1) {
		if (sxsi_hd[i].fname[0]) {
			MEM_DISK_EQUIP |= bit;
		}
	}
}

static void init_scsi_equip(void) {

	BYTE	i;
	BYTE	bit;
	WORD	w;

	MEM_DISK_EQUIPS = 0;
	ZeroMemory(&mem[0x00460], 0x20);

	for (i=0, bit=1; i<2; i++, bit<<=1) {
		if (sxsi_hd[i+2].fname[0]) {
			MEM_DISK_EQUIPS |= bit;
			mem[0x00460+i*4] = sxsi_hd[i+2].sectors;
			mem[0x00461+i*4] = sxsi_hd[i+2].surfaces;
			switch(sxsi_hd[i+2].size) {
				case 256:
					w = 0 << 12;
					break;
				case 512:
					w = 1 << 12;
					break;
				default:
					w = 2 << 12;
					break;
			}
			*(WORD *)(&mem[0x00462+i*4]) = (w | sxsi_hd[i+2].tracks);
		}
	}
}


static BYTE sxsi_pos(long *pos) {

	SXSI_TABLE	*sxsi;
	int			np2drv;

	*pos = 0;
	np2drv = (I286_AL & 0x20) >> 4;
	if ((I286_AL & 0x0f) >= 2) {
		return(0x60);
	}
	np2drv |= (I286_AL & 1);
	sxsi = &sxsi_hd[np2drv];
	if (I286_AL & 0x80) {
		if ((I286_DL >= sxsi->sectors) ||
			(I286_DH >= sxsi->surfaces) ||
			(I286_CX >= sxsi->tracks)) {
			return(0xd0);
		}
		(*pos) = ((I286_CX * sxsi->surfaces) + I286_DH) * sxsi->sectors
															+ I286_DL;
	}
	else {
		*pos = (I286_DL << 16) | I286_CX;
		if (!(I286_AL & 0x20)) {
			(*pos) &= 0x1fffff;
		}
		if ((*pos) >= sxsi->totals) {
			return(0xd0);
		}
	}
	return(0x00);
}

static BYTE sxsi_read(BYTE drv, long pos, BYTE *dest, WORD size) {

	FILEH	hdh;
	long	r;
	WORD	rsize;
	int		np2drv;

	np2drv = (drv & 0x20) >> 4;
	if ((drv & 0x0f) >= 2) {
		return(0x60);
	}
	np2drv |= (drv & 1);
	if ((!sxsi_hd[np2drv].fname[0]) ||
		((hdh = file_open(sxsi_hd[np2drv].fname)) == (FILEH)-1)) {
		return(0x60);
	}
	pos = pos * sxsi_hd[np2drv].size + sxsi_hd[np2drv].headersize;
	r = file_seek(hdh, pos, FSEEK_SET);
	if (r == -1) {
		file_close(hdh);
		return(0x40);
	}
	if (pos != r) {
		file_close(hdh);
		return(0xd0);
	}
	while(size) {
		if (size >= sxsi_hd[np2drv].size) {
			rsize = sxsi_hd[np2drv].size;
		}
		else {
			rsize = size;
		}
		if (file_read(hdh, dest, rsize) != rsize) {
			file_close(hdh);
			return(0xd0);
		}
		dest += rsize;
		size -= rsize;
	}
	file_close(hdh);
	return(0x00);
}


static BYTE sxsi_write(BYTE drv, long pos, BYTE *src, WORD size) {

	FILEH	hdh;
	long	r;
	WORD	wsize;
	int		np2drv;

	np2drv = (drv & 0x20) >> 4;
	if ((drv & 0x0f) >= 2) {
		return(0x60);
	}
	np2drv |= (drv & 1);
	if ((!sxsi_hd[np2drv].fname[0]) ||
		((hdh = file_open(sxsi_hd[np2drv].fname)) == (FILEH)-1)) {
		return(0x60);
	}
	pos = pos * sxsi_hd[np2drv].size + sxsi_hd[np2drv].headersize;
	r = file_seek(hdh, pos, FSEEK_SET);
	if (r == -1) {
		file_close(hdh);
		return(0x40);
	}
	if (pos != r) {
		file_close(hdh);
		return(0xd0);
	}
	while(size) {
		if (size >= sxsi_hd[np2drv].size) {
			wsize = sxsi_hd[np2drv].size;
		}
		else {
			wsize = size;
		}
		if (file_write(hdh, src, wsize) != wsize) {
			file_close(hdh);
			return(0x70);
		}
		src += wsize;
		size -= wsize;
	}
	file_close(hdh);
	return(0x00);
}

static BYTE sxsi_operate(BYTE type) {

	BYTE	ret_ah = 0x00;
	BYTE	drv;
	int		i;
	long	pos;

	drv = (I286_AL & 0x20) >> 4;
	if ((I286_AL & 0x0f) >= 2) {
		return(0x60);
	}
	drv |= (I286_AL & 1);

	switch(I286_AH & 0x0f) {
		case 0x01:						// xt@C
		case 0x07:						// gNg
		case 0x0e:						// [hZbg
		case 0x0f:						// gNg
			break;

   		case 0x03:						// CjVCY
			if (type == HDDTYPE_SASI) {
				init_sasi_equip();
			}
			else if (type == HDDTYPE_SCSI) {
				init_scsi_equip();
			}
			break;

   		case 0x04:						// ZX
			ret_ah = 0x00;
			if ((I286_AH == 0x04) && (type == HDDTYPE_SASI)) {
				ret_ah = 0x04;
			}
			else if ((I286_AH == 0x44) && (type == HDDTYPE_SCSI)) {
				I286_BX = 1;
			}
			else if (I286_AH == 0x84 || I286_AH == 0x80) {
				I286_BX = sxsi_hd[drv].size;
				I286_CX = sxsi_hd[drv].tracks;
				I286_DH = sxsi_hd[drv].surfaces;
				I286_DL = sxsi_hd[drv].sectors;
			}
			break;

		case 0x05:						// f[^̏
			i286_memstr_read(I286_ES, I286_BP, work, I286_BX);
			ret_ah = sxsi_pos(&pos);
			if (!ret_ah) {
				ret_ah = sxsi_write(I286_AL, pos, work, I286_BX);
			}
			break;

		case 0x06:						// f[^̓ǂݍ
			ret_ah = sxsi_pos(&pos);
			if (!ret_ah) {
				ret_ah = sxsi_read(I286_AL, pos, work, I286_BX);
				if (ret_ah < 0x20) {
					i286_memstr_write(I286_ES, I286_BP, work, I286_BX);
				}
			}
			break;

		case 0x0d:						// tH[}bg
			if (I286_DL) {
				ret_ah = 0x30;
				break;
			}
			i286_memstr_read(I286_ES, I286_BP, work, I286_BX);
			ret_ah = sxsi_pos(&pos);
			memset(work, 0xe5, sxsi_hd[drv].size);
			for (i=0; (i<sxsi_hd[drv].sectors) && (!ret_ah); i++) {
				ret_ah = sxsi_write(I286_AL, pos++, work, sxsi_hd[drv].size);
			}
			break;

		default:
			ret_ah = 0x40;
			break;
	}
	return(ret_ah);
}

// -------------------------------------------------------------------- BIOS

void bios0x1b_hddreopen(void) {

	int				i;
	WORD			tracks;
	FILEH			hdh;
	char			*ext;
	SXSI_TABLE		*sxsi;
	int				sasi = 0;
	int				scsi = 0;
//	V98SCSI_TABLE	v98;

	ZeroMemory(sxsi_hd, sizeof(sxsi_hd));
	for (i=0; i<2; i++) {
		if (hd_name[i][0]) {
			ext = milstr_getext(hd_name[i]);
			if (!milstr_cmpA(ext, "THD")) {			// T98 HDD (SASI)
				hdh = file_open(hd_name[i]);
				if (hdh != (FILEH)-1) {
					sxsi = &sxsi_hd[sasi];
					*sxsi = defsasi;
					if (file_read(hdh, &tracks, 2) == 2) {
						sxsi->tracks = tracks;
						sxsi->totals = sxsi->tracks *
											sxsi->sectors * sxsi->surfaces;
						strcpy(sxsi->fname, hd_name[i]);
						sasi++;
					}
					file_close(hdh);
				}
			} else if (!milstr_cmpA(ext, "HDI")) {	// ANEX86 HDD (SASI)
				hdh = file_open(hd_name[i]);
				if (hdh != (FILEH)-1) {
					HDI_HEADER hdihead;
					sxsi = &sxsi_hd[sasi];
					if (file_read(hdh, &hdihead, 32) == 32) {
						sxsi->size = hdihead.sector_size;
						sxsi->headersize = hdihead.header_size;
						sxsi->tracks = hdihead.track;
						sxsi->surfaces = hdihead.head;
						sxsi->sectors = hdihead.sector;
						sxsi->totals = sxsi->tracks *
											sxsi->sectors * sxsi->surfaces;
						strcpy(sxsi->fname, hd_name[i]);
						sasi++;
					}
					file_close(hdh);
				}
			}
#if 0
			else if (!milstr_cmpA(ext, "HDD")) {	// Virtual98 HDD (SCSI)
				hdh = file_open(hd_name[i]);
				if (hdh != (FILEH)-1) {
					if ((file_read(hdh, &v98, sizeof(v98)) == sizeof(v98)) &&
						(!memcmp(v98.vhd, "VHD", 3))) {
						sxsi = &sxsi_hd[scsi+2];
						*sxsi = defscsi;
						sxsi->totals = v98.totals;
						sxsi->tracks = v98.tracks;
						sxsi->size = v98.sectorsize;
						sxsi->sectors = v98.sectors;
						sxsi->surfaces = v98.surfaces;
						strcpy(sxsi->fname, hd_name[i]);
						scsi++;
					}
					file_close(hdh);
				}
			}
#endif
		}
	}
}

void bios0x1b_allclose(void) { }

BYTE boot_fd(BYTE drv) {

	WORD	remain;
	WORD	size;
	BYTE	mediatype;
	DWORD	pos;

	if (drv >= 4) {
		return(0xd0);
	}
	fdc.us = drv & 3;
	if (!fdd_diskready(fdc.us)) {
		return(0x60);
	}
	for (mediatype=0; mediatype<3; mediatype++) {
		CTRL_FDMEDIA = mediatype;
		if (!biosfd_seek(0)) {
			fdc.hd = 0;
			if (fdd_readid()) {
				return(0xa0);
			}
			remain = 0x400;
			pos = 0x1fc00;
			while(remain) {
				if (fdd_read()) {
					return(0xc0);
				}
				if (fdc.N < 3) {
					size = 128 << fdc.N;
				}
				else {
					size = 128 << 3;
				}
				if (remain < size) {
					memcpy(&mem[pos], fdc.buf, remain);
					break;
				}
				else {
					memcpy(&mem[pos], fdc.buf, size);
					pos += size;
					remain -= size;
					fdc.R++;
				}
			}
			switch(mediatype) {
				case DISKTYPE_2D:
					MEM_DISK_BOOT = (BYTE)(0x50+drv);
					fdmode = 3;
					break;
				case DISKTYPE_2DD:
					MEM_DISK_BOOT = (BYTE)(0x10+drv);
					fdmode = 3;
					break;
				default:
					MEM_DISK_BOOT = (BYTE)(0x90+drv);
					break;
			}
			return(0x00);
		}
	}
	return(0xc0);
}


BYTE boot_hd(BYTE drv) {

	BYTE	ret;

	ret = sxsi_read(drv, 0, &mem[0x1fc00], 0x400);
	if (ret < 0x20) {
		MEM_DISK_BOOT = drv;
	}
	return(ret);
}



void bootstrapload(void) {

	BYTE	i;
	BYTE	bootret;

	fdmode = 0;
	bootret = 0xd0;
	for (i=0; (i<4) && (bootret >= 0x20); i++) {
		if (fdd_diskready(i)) {
			bootret = boot_fd(i);
		}
	}
	for (i=0; (i<2) && (bootret >= 0x20); i++) {
		if (sxsi_hd[i+2].fname[0]) {
			bootret = boot_hd(i | 0xa0);
		}
	}
	for (i=0; (i<2) && (bootret >= 0x20); i++) {
		if (sxsi_hd[i].fname[0]) {
			bootret = boot_hd(i | 0x80);
		}
	}
	init_fdd_equip();
	init_sasi_equip();
	init_scsi_equip();
	if (bootret >= 0x20) {
		CopyMemory(&mem[0x1fc00], nosyscode, sizeof(nosyscode));
	}
}

// --------------------------------------------------------------------------

void bios0x1b(void) {

	BYTE	ret_ah;
	BYTE	flag;

	switch(I286_AL & 0xf0) {
		case 0x90:
			ret_ah = fdd_operate(DISKTYPE_2HD);
			break;

		case 0x10:
		case 0x70:
		case 0xf0:
			ret_ah = fdd_operate(DISKTYPE_2DD);
			break;

		case 0x50:
			ret_ah = fdd_operate(DISKTYPE_2D);
			break;

		case 0x00:
		case 0x80:
			ret_ah = sxsi_operate(HDDTYPE_SASI);
			break;
#if 0
		case 0x20:
		case 0xa0:
			ret_ah = sxsi_operate(HDDTYPE_SCSI);
			break;
#endif

		default:
			ret_ah = 0x40;
			break;
	}
#if 0 //ifdef TRACE
	{
		char	buf[256];
		wsprintf(buf, "%04x:%04x AX=%04x BX=%04x %02x:%02x:%02x:%02x\n"	\
						"ES=%04x BP=%04x \nret=%02x",
			i286_memword_read(I286_SS, I286_SP+2),
			i286_memword_read(I286_SS, I286_SP),
			I286_AX, I286_BX, I286_CL, I286_DH, I286_DL, I286_CH, 
			I286_ES, I286_BP, ret_ah);
			TRACE_(buf, 0);
	}
#endif

	I286_AH = ret_ah;
	flag = i286_membyte_read(I286_SS, I286_SP+4) & 0xfe;
	if (ret_ah >= 0x20) {
		flag |= 1;
	}
	i286_membyte_write(I286_SS, I286_SP+4, flag);
}
